/*
 * The codes simulate a single thread of simulation for given initial long&short bursts numbers.
 */
 
package PLBP;

/**
 *
 * @author Chen Jiongze
 */
public class PLBPSimInstance {
    private double T;    
    private final int intervalNo;
    private final double intervalLength;
    private final int longBurstNo;
    private final int shortBurstNo;
    private double[] works; //works of each interval
    
    /**
     * 
     * @param T total simulation time
     * @param intervalNo number of intervals
     * @param longBurstNo number of initial long bursts
     * @param shortBurstNo number of initial short bursts
     */
    public PLBPSimInstance(double T, int intervalNo, int longBurstNo,
            int shortBurstNo){
        this.T = T;
        this.longBurstNo = longBurstNo;
        this.shortBurstNo = shortBurstNo;
        this.intervalNo = intervalNo;
        
        //obtain the number of time intervals
        intervalLength = T/(double)intervalNo;   
        
        //initial the array of works at all time intervals
        works = new double[intervalNo];
        for(int i = 0; i < intervalNo; i++){
            works[i] = 0.0;
        }
    }
    
    /**
     * 
     * @param thresholds a double array containing the buffer threshold, x
     * @param lambda arrival rate lambda    
     * @param gamma shape parameter of the Lomax distribution
     * @param delta scale parameter of the Lomax distribution
     * @param r the constant rate at which each burst generate work 
     * @param C capacity/bandwidth
     * @return an array of overflow probabilities, P(Q>x), for all x
     * values in the thresholds
     */
    public double[] runForOverflowProb(double[] thresholds, double lambda,
            double gamma, double delta, double r, double C){      
        double currentTime = RVGenerator.getExpRV(1.0/lambda);
        //index of current time interval
        int checkPointNo = 1;
        //the work left in the queue at the end of the interval
        double workLeft = 0.0;
        //array of countors corresponding to x values
        int[] counts = new int[thresholds.length];  
        for(int i = 0; i < counts.length; i++){
            counts[i] = 0;
        }

        //adding the initial long and short bursts into the queue
        initialQueue(gamma, delta, r);
        
        //start simulation
        while(currentTime < T){     //simulation does not end
            while(currentTime >= (double)checkPointNo * intervalLength){ 
                //a interval ends, start to monitor the queue
                workLeft = Math.max(workLeft + works[checkPointNo-1]
                        - C*intervalLength,0.0);
                //counter++ for the corresponding x if the queue size >= x
                int tmpPos = searchPosition(thresholds, 0, workLeft);
                for(int i = 0; i < tmpPos; i++){
                    counts[i]++;
                }
                checkPointNo++;
            }
            //assign work to the intervals
            double duration = RVGenerator.getLomaxRV(gamma, delta);
            double departureTime = currentTime + duration;
            addWork(checkPointNo, currentTime, departureTime, duration, r);
            
            currentTime += RVGenerator.getExpRV(1.0/lambda);
        }
        
        //monitor the queue at the end of the simulation
        int tmpPos = searchPosition(thresholds, 0, Math.max(workLeft +
                works[checkPointNo-1] - C*intervalLength,0.0));
        for(int i = 0; i < tmpPos; i++){
            counts[i]++;
        }
        
        //calculate the overflow probabilities
        double[] results = new double[counts.length+1];
        for(int i = 0; i < counts.length; i++){
            results[i] = (double)counts[i] / (double)intervalNo;
        }
        double mean = 0.0;
        for(int i = 0; i < works.length; i++){
            mean += works[i];
        }
        mean =  mean / intervalLength / (double) works.length;
        
        //add mean of the workloads for one interval to the array
        results[counts.length] = mean;
        return results;
    }
    
    /**
     * This function initializes the queue by adding the
     * initial long and short bursts into the queue
     * @param gamma shape parameter of the Lomax distribution
     * @param delta scale parameter of the Lomax distribution
     * @param r the constant rate at which each burst generate work
     */
    private void initialQueue(double gamma, double delta, double r){
        double duration;
        for(int i = 0; i < works.length; i++){  //for long burst
            works[i] += (double)longBurstNo * r * intervalLength;
        }
        for(int i = 0; i < shortBurstNo; i++){ //for short burst
            duration = initialShortBurst(gamma, delta);
            addWork(1, 0, duration, duration, r);
        }
    }
    
    /**
     * 
     * @param checkPointNo index of the current interval,
     * belonging to [1, intervalLength]
     * @param arrivalTime arrival time of the burst
     * @param departureTime departure time of the burst
     * @param duration the duration of the burst
     * @param r the constant rate at which each burst generate work
     */
    private void addWork(int checkPointNo, double arrivalTime, 
            double departureTime, double duration, double r){
        if(departureTime <= (double)checkPointNo * intervalLength){
            //end before current interval
            works[checkPointNo-1] += duration * r;
        }else{  //end after current interval
            if(checkPointNo == intervalNo){ //last interval
                works[checkPointNo-1] += (T - arrivalTime) * r;
            }else{  //not last interval
                works[checkPointNo-1] += ((double)checkPointNo 
                        * intervalLength - arrivalTime) * r;
                if(works[checkPointNo-1]<0){
                        System.out.println("error");
                    }
                int no = checkPointNo+1;
                while(departureTime > (double)no * intervalLength
                        && no < intervalNo){ //end after "no" interval
                    works[no-1] += intervalLength * r;
                    no++;
                    if(works[no-1]<0){
                        System.out.println("error");
                    }
                }
                if(departureTime > T){
                    works[no-1] += intervalLength * r;
                }else{
                    works[no-1] += (departureTime - (double)(no - 1)
                            * intervalLength) * r;
                }
                if(works[no-1]<0){
                        System.out.println("error");
                    }
            }
        }
    }
    
    /**
     * 
     * @param array input double array  
     * @param start the start searching position
     * @param x value for comparison
     * @return the first pos >= start, that makes array[pos] >= x;
     * or array.length if all values in array is less than x
     */
    private int searchPosition(double[] array, int start, double x){
        int pos = start;
        while(pos < array.length && array[pos] < x){
            pos++;
        }
        return pos;
    }
    
    /**
     * 
     * @param gamma shape parameter of the Lomax distribution
     * @param delta scale parameter of the Lomax distribution
     * @return the duration of an initial short burst
     */
    private double initialShortBurst(double gamma, double delta){
        double burst = RVGenerator.getFRTLomaxRV(gamma, delta);
        while(burst > T){
            //the duration is longer than total simulation time
            burst = RVGenerator.getFRTLomaxRV(gamma, delta);
        }
        return burst;
    }    
}
