/*
 * This codes are the fast simulation method provided by the paper:
 * Performance Evaluation of a Queue Fed by a Poisson Lomax Burst Process.
 */
package PLBP;

import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import javax.swing.JOptionPane;

/**
 *
 * @author Chen Jiongze
 */
public class PLBPSimulation {
    private double[] thresholds;
    private double lambda, gamma, delta, r, C;
    private double trafficMean;
    public PLBPSimulation(double[] thresholds, double lambda, double gamma, double delta, double r, double C){
        this.thresholds = thresholds;
        this.lambda = lambda;
        this.gamma = gamma;
        this.delta = delta;
        this.r = r;
        this.C = C;
        trafficMean = 0.0;
    }
    
    /**
     * Fast simulation method to obtain the overflow probabilities for PLBP queues.
     * @param T total simulation time for a single thread of simulation
     * @param intervalNo no. of intervals the T is divided into
     * @param iterations number of iterations of the threads for a given initial states
     * @return array of overflow probabilities
     */
    public double[] runSim(double T, int intervalNo, int iterations){
        String info;
        double durationMean = PLBPStats.getLomaxMean(gamma, delta);
        double totalMean = PLBPStats.getMean(1, lambda, gamma, delta, r);
        double beta = PLBPSimStats.getBeta(T, lambda, gamma, delta, r, C);
        double shortBurstNoMean = lambda * durationMean - beta;
        
        info = "\nPLBP Queue\nlambda = "+ lambda + ", gamma = " + gamma + ", delta = " + delta + ", r = " + r + ", C = " + C
                + "\nT = " + T + ", interval no. = " + intervalNo + ", iteration no = " + iterations + 
                ", intervalLength =" + T/(double)intervalNo + ".\n";
        info += "totalMean = " + totalMean +", durationMean = " + durationMean
                + "\nbeta = " + beta +", initial shortBurstNoMean = " + shortBurstNoMean;
        System.out.println(info);
        
        Date date = new Date();
        System.out.println("Simulation starts at " + date.toString());
        double[] probs = getOverflowProb(T, intervalNo, iterations, beta, shortBurstNoMean);
        date = new Date();
        System.out.println("Simulation ends at " + date.toString());
        return probs;
    }
    
    //initial long burst number = 0, 1, 2, 3, 4 ... size
    private double[] getOverflowProb(double T, int intervalNo, int iterations, double beta, double shortBurstNoMean){
        int thre = iterations/10;
        int size = getNSize(beta);
        System.out.println("n in Tim's code = "+PLBPSimStats.getPhi(T, lambda, gamma, delta, r, C));
        System.out.println("n stops at "+size+", expceted short burst mean no = "+shortBurstNoMean);
        double[][] probs = new double[size][thresholds.length];
        PLBPSimInstance instance;
        for(int i = 0; i < size; i++){
            for(int j = 0; j < thresholds.length; j++){
                probs[i][j] = 0;
            }
        } 
        int longBurstNo = 0, shortBurstNo;
        double[] tmpProbs;
        double[] means = new double[size];
        for(int i = 0; i < size; i++){           
            longBurstNo = i;
            means[i] = 0.0;
            System.out.println("Initial Long Burst no. = "+i+" ...");
            double shortBurstMeanNo = 0;
            for(int j = 1; j < iterations+1; j++){               
                shortBurstNo = RVGenerator.getPoissonRV(shortBurstNoMean);
                shortBurstMeanNo += shortBurstNo;
                instance = new PLBPSimInstance(T, intervalNo, longBurstNo, shortBurstNo);
                tmpProbs = instance.runForOverflowProb(thresholds, lambda, gamma, delta, r, C);
                for(int z = 0; z < tmpProbs.length - 1; z++){
                    probs[i][z] = (tmpProbs[z] + (double)(j-1)*probs[i][z])/(double)j;
                } 
                means[i] = ((double)(j-1) * means[i] + tmpProbs[tmpProbs.length - 1])/(double)j;  
                if(thre==0){
                    System.out.print(" ["+j+"th it.]");
                }else if(j%thre==0){
                    System.out.print(" ["+10*(int)(j/thre)+"%]");
                }
            }
            System.out.println();
            System.out.println("\t Finished. Short burst mean no ="+shortBurstMeanNo/(double)iterations+", PLBP mean = "+ means[i]+".");
        }
        double[] finalProbs = new double[thresholds.length];
        for(int i = 0; i < thresholds.length; i++){
            finalProbs[i] = 0;
            for(int j = 0; j < size; j++){
                finalProbs[i] += getPoissonProb(j, beta) * probs[j][i];
            }
        }

        for(int i = 0; i < size; i++){
                trafficMean += getPoissonProb(i, beta) * means[i];
        }
        System.out.println("Final PLBP Mean = "+ trafficMean+".");
        return finalProbs;
    }
    
        
    private double getPoissonProb(int n, double lambda){
        if(n == 0){
            return Math.exp(-lambda);
        }
        return getPoissonProb(n-1, lambda) * lambda / (double) n;
    }
    
    private int getNSize(double lambda){
        int n = (int)Math.floor(lambda);
        int limit = (int)Math.floor(C/r);
        while (getPoissonProb(n, lambda)>0.000000001 && n <= limit){
            n++;
        }
        return n+1;
    }
    

    private double getMean(double[] data){
        double sum = 0.0;
        for(double a : data){
            sum += a;
        }
        return sum/(double)data.length;
    }

    private double getVariance(double[] data, double mean){
        double temp = 0.0;
        for(double a :data){
            temp += (mean-a)*(mean-a);
        }
        return temp/data.length;
    }
}
