
#include "qsform.h"

/* Author: Jiongze Chen & Tim Neame
 * Date: 10/08/2015
 * The codes are developed for PLBP queue based on Tim's codes for PPBP
 * for paper: Performance Evaluation of a Queue Fed by a Poisson Lomax Burst Process.
 * The codes modefied by Jiongze are with comments, change.
 *
 * The idea here is to use the AMN formula to estimate the SRD queueing
 * performance, then combine the results for different numbers of long
 * bursts.
 * How do we decide what time period to use to differentiate between the SRD
 * process and the long bursts? The answer is that we don't, we try a 
 * range of values of W, and choose the highest loss probability
 * We also need to have a modified variance function for the SRD process.
 */

/* Calculate the probability than a Poisson distribution with given mean
 * will produce a sample of n */
 
double CalculatePoisson(double mean, int n)
{
  double value = exp(-mean);

  for(int i = 1; i <= n; i++)
    value *= (mean/(double)i);

  return value;
}
QuasiFormula::QuasiFormula()
{
  lambda = 0.0;
  r = 0.0;
  delta = 0.0;
  Gamma = 0.0;
  kappa = 0.0;
  Lomaxmean = 1.0;	//change: Paretomean to Lomaxmean
  noCBRmean = -1.0;
  mean = 1.0;
  variance = 1.0; 
}

/* Set up a QuasiFormula object with defined values of lambda, r, delta,
 * gamma and kappa.
 */

QuasiFormula::QuasiFormula(double l, double rVal, double d, double g, double k)
{
  lambda = l;
  r = rVal;
  delta = d;
  Gamma = g;
  kappa = k;
  Lomaxmean = delta /(Gamma - 1.0);	//change: Paretomean to Lomaxmean, Paretomean = (delta*Gamma)/(Gamma - 1.0)
  noCBRmean = r*lambda*Lomaxmean;	//change: Paretomean to Lomaxmean
  mean = noCBRmean + kappa;
  variance = Calcvt(1.0); 
}

QuasiFormula::~QuasiFormula()
{
}

double QuasiFormula::GetMean() const
{
  return mean;
}

double QuasiFormula::GetNoCBRMean() const
{
  return noCBRmean;
}

//change to the variance function for PLBP
/* Calculate the value of the variance for a given value of t
 */
double QuasiFormula::Calcvt(double t)
{
  double c1 = 2.0 * lambda * r * r * delta * delta * delta /(Gamma - 1.0)/(Gamma - 2.0);
  double c2 = t/delta + (1.0 - pow(1 + t/delta, 3.0 - Gamma))/(3.0 - Gamma);
  return c1 * c2;
}

//change to the variance of short burst process of PLBP
/* Calculate the modified variance value for a value of t. The modified
 * variance will be the value used in calculating the short bursts value.
 */

double QuasiFormula::CalcModvt(double t, double W)
{
  double modVar = t*t*r*r*lambda*delta*pow(1.0+W/delta,(1.0-Gamma))/(Gamma - 1.0);
  double tmpVar = Calcvt(t);
  return (tmpVar - modVar);

}

//change to the first derivate of the variance of PLBP
/* The AMN formula uses the first derivate (wrt to t) of tha variance
 * function. This function calculate the value of that first derivative
 * at a given value of t.
 */

double QuasiFormula::Calcvprime(double t)
{
	double c1 = 2.0 * lambda * r * r * delta * delta * delta /(Gamma - 1.0)/(Gamma - 2.0);
	double c2 = 1.0/delta - pow(1 + t/delta,2.0 - Gamma)/delta;
	return c1 * c2;
}

//change to the first derivate of the variance of short burst process of PLBP
/* Just as the value of the variance is modified in calculating the short
 * bursts component value, so also must the value of its derivative be
 * modified. This function calculate the modified value.
 */

double QuasiFormula::CalcModvprime(double t, double W)
{
  double modVarPrime = 2.0*t*r*r*lambda*delta*pow(1.0+W/delta,(1.0-Gamma))/(Gamma - 1.0);
  double tmpVar = Calcvprime(t);
  return (tmpVar - modVarPrime);
}

/* To calculate the AMN estimate, the value t* (tStar) must first be 
 * identified.
 * x is actually x/(c-m) in the formula (4.20)
 */

double QuasiFormula::FindtStar(double x, double W)
{
  const double MinAccuracy = 1e-6;
  int CurrentLHSIsLess, LastLHSIsLess;

  double vt = 0.00;
  double vprime = 0.0;
  double lower, upper, middle;

  // Find t to the nearest integer value.
  // Note that there are assumptions here that:
  // 1. The value of t > 0.
  // 2. The function ((2*vt)/vprime) - t is increasing.
  double t = 0.00;

  do
    {
    t += 1.0;
    vt = CalcModvt(t, W);

    vprime = CalcModvprime(t, W);

    } while(((((2*vt)/vprime) - t) < x) && (t <= W));

  if(t > W){
    //return t;
	return W;
  }

  // Now refine the accuracy of the answer using a binary search
  // technique.
  lower = t - 1.0;
  upper = t;

  do
  {
    middle = (upper + lower)/2.0;

    vt = CalcModvt(middle, W);
    vprime = CalcModvprime(middle, W);
    if((((2*vt)/vprime) - middle) < x)
      CurrentLHSIsLess = True;
    else
      CurrentLHSIsLess = False;

    if(CurrentLHSIsLess == True)
      lower = middle;
    else
      upper = middle;
  } while(fabs((upper - lower)/lower) > MinAccuracy);

  return middle;

}

/* Calculate the AMN estimate of the queueing performance of a given
 * short bursts process.
 */

double QuasiFormula::CalcAMNEstimate(double t, double x, double m, 
		double c, double W)
{
  double vt = CalcModvt(t, W);
  double tmpResult = exp((-(x + (c-m)*t)*(x + (c-m)*t))/(2.0*vt));

  return tmpResult;
}

double QuasiFormula::CalcQSForW(double W, double x, double servrate)
{
  double netSR,valn,tStar,ScaleFactor;
  double LongBurstMean, ShortBurstMean;
  double sum = 0.0;
  
  LongBurstMean = lambda*delta /(Gamma- 1.0) * pow(1+W/delta,(1.0-Gamma));	//change the mean of long burst process for PLBP
  ShortBurstMean = noCBRmean/r - LongBurstMean;
  for(int n = 0; n <= int((servrate - mean)/r) ; n++)
    {
    // Net sevice rate = original capacity, minus the capacity used 
    // by the n long bursts, minus the mean rate of the short burst process.
    // Original capacity = servrate - (noCBRmean + kappa)
    netSR = (servrate - kappa) - n*r - ShortBurstMean*r;

    ScaleFactor = 1.0;

    tStar = FindtStar((x/netSR), W);
    if(tStar > W)
      {
      tStar = W;
      }
    valn = ScaleFactor*CalcAMNEstimate(tStar,x, 0.0, netSR, W);

    // Poisson probability that there will be n long bursts
    sum += (CalculatePoisson(LongBurstMean,n)*valn);
    }

  return sum;
}

/* Taken from Numerical Recipes, Section 10.1, except that I have
 * modified the routine to find a maximum, not a minimum.
 * Given a function, func and initial points aW and bW, this function
 * returns points aW, bW, cW that bracket a maximum of the function.
 * Also returns the function values at those points, lossa, lossb, lossc */

void QuasiFormula::MaxQSBracket(double *aW, double *bW, double *cW,
	double *lossa, double *lossb, double *lossc,
	double x, double servrate)
{
  double ulim,u,r,q,lossu,dum;

  *lossa = CalcQSForW(*aW, x, servrate);
  *lossb = CalcQSForW(*bW, x, servrate);

  if(*lossb < *lossa)
    /* Switch a and b so we can go uphill in the direction from a to b */
    {
      SHFT(dum,*aW,*bW,dum)
      SHFT(dum,*lossa,*lossb,dum)
    }

  *cW = (*bW)+GOLD*(*bW-*aW);
  *lossc = CalcQSForW(*cW, x, servrate);

  while(*lossb < *lossc)
    {
    r = (*bW-*aW)*(*lossb-*lossc);
    q = (*bW-*cW)*(*lossb-*lossa);
    u = (*bW) - ((*bW-*cW)*q-(*bW-*aW)*r/
	2.0*SIGN(DMAX(fabs(q-r),TINY),q-r));
    ulim = (*bW)+GLIMIT*(*cW-*bW);

    if((*bW-u)*(u-*cW) > 0.0)
      /* u is between b and c, so try it */
      {
      lossu = CalcQSForW(u, x, servrate);

      if(lossu > *lossc)
	/* Got a maximum between b and c */
	{
	*aW = (*bW);
	*bW = u;
	*lossa = (*lossb);
	*lossb = lossu;
	return;
	}
      else if(lossu < *lossb)
	/* Got a maximum between a and u
	{
	*cW = u;
	*lossc = lossu;
	return;
	}

      /* Parabolic fit was no use. Use default magnification */
      u = (*cW) + GOLD*(*cW-*bW);
      lossu = CalcQSForW(u, x, servrate);
      }
    else if((*cW - u)*(u-ulim) > 0.0)
      /* Parabolic fit is between c and its allowed limit. */
      {
      lossu = CalcQSForW(u, x, servrate);
      if(lossu < (*lossc))
	{
	SHFT(*bW,*cW,u,*cW+GOLD*(*cW-*bW))
	SHFT(*lossb,*lossc,lossu,CalcQSForW(u, x, servrate));
	}
      }
    else if((u-ulim)*(ulim-*cW) >= 0.0)
      /* Limit parabolic u to maximum allowed value. */
      {
      u = ulim;
      lossu = CalcQSForW(u, x, servrate);
      }
    else
      /* Reject parabolic u, use default magnification */
      {
      u = (*cW)+GOLD*(*cW-*bW);
      lossu = CalcQSForW(u, x, servrate);
      }

    SHFT(*aW,*bW,*cW,u)
    SHFT(*lossa,*lossb,*lossc,lossu)
    }
}

/* Given three bracketing values for W (aW, bW and cW) s.t.
 * bx lies between aW and cW, and f(bW) is greater than both
 * f(aW) and f(bW) this routine finds the maximum value. The maximum
 * value of the function is returned and the corresponding W
 * value is passed back as Wmax. 
 */

double QuasiFormula::GoldenQS(double aW, double bW, double cW, 
	double x, double servrate, double *Wmax)
{
  double f1,f2,W0,W1,W2,W3;

  W0 = aW;
  W3 = cW;

  if(fabs(cW-bW) > fabs(bW-aW))
    // Make W0 to W1 the smaller segment
    {
    W1 = bW;
    W2 = bW + GOLDC*(cW-bW);
    }
  else
    {
    W2 = bW;
    W1 = bW - GOLDC*(bW-aW);
    }

  f1 = CalcQSForW(W1, x, servrate);
  f2 = CalcQSForW(W2, x, servrate);

  while(fabs(W3-W0) > MinAccuracy*(fabs(W1)+fabs(W2)))
    {
    // cout << W0 << "," << W1 << "," << W2 << "," << W3 << endl;
    if(f2 > f1)
      {
      SHFT(W0,W1,W2,R*W1+GOLDC*W3);
      SHFT2(f1,f2,CalcQSForW(W2, x, servrate));
      }
    else
      {
      SHFT(W3,W2,W1,R*W2+GOLDC*W0);
      SHFT2(f2,f1,CalcQSForW(W1, x, servrate));
      }
    }
  
   //cout << W0 << "," << W1 << "," << W2 << "," << W3 << endl;
  if(f1 > f2)
    {
    *Wmax = W1;
	CalcQSForW(W1, x, servrate);
	//cout << *Wmax << "," << W1 << endl;
    return f1;
    }
  else
    {
    *Wmax = W2;
	CalcQSForW(W2, x, servrate);
	//cout << *Wmax << "," << W2 << endl;
    return f2;
    }
}

double QuasiFormula::CalcQSEstimate(double x, double servrate,
		double *maxW)
{
  double maxloss = 0.0;
  double val;

  double loss1, loss2, loss3;
  double startWmin, startWmax, startWmid;
  startWmin = 1.1;
  startWmax = 8000.0;
  startWmid = 100.0;
  
  *maxW = 0.0;
  if(x > 0.0)
    {
    MaxQSBracket(&startWmin, &startWmid, &startWmax,
	&loss1, &loss2, &loss3, x, servrate);
    maxloss = GoldenQS(startWmin, startWmid, startWmax, x, servrate, maxW);
    }
  else
    {
    // A zero value for the buffer threshold causes problems.
    // Move the value a little, to get a reasonable estimate.
    MaxQSBracket(&startWmin, &startWmid, &startWmax,
	&loss1, &loss2, &loss3, 0.01, servrate);
    maxloss = GoldenQS(startWmin, startWmid, startWmax, 0.01, servrate,
	maxW);
    }

  return maxloss;
}

// Display the values of the parameters set inside this object.
 
ostream& QuasiFormula::Show(ostream& out)
{
  out << "delta = " << delta << endl;
  out << "gamma = " << Gamma << endl;
  out << "Mean burst size = " << Lomaxmean << endl;
  out << "lambda = " << lambda << endl;
  out << "r = " << r << endl << endl;
  out << "CBR component = " << kappa << endl << endl;
  out << "Mean without CBR = " << noCBRmean << endl;
  out << "mean = " << mean << endl;
  out << "variance = " << variance << endl;
  out << "H = " << (3.0-Gamma)/2.0 << endl;
  return out;
}

ostream& operator<<(ostream& out, QuasiFormula& qf)
{
  qf.Show(out);
  return out;
}

